﻿{
  Jpeg2kBitmap.pas for OpenJpeg21.pas

  The contents of this file are used with permission, subject to the Mozilla
  Public License Version 1.1 (the "License"); you may not use this file except
  in compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/MPL-1.1.html

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
  the specific language governing rights and limitations under the License.

  Alternatively, the contents of this file may be used under the terms of the
  GNU Lesser General Public License (the  "LGPL License"), in which case the
  provisions of the LGPL License are applicable instead of those above.
  If you wish to allow use of your version of this file only under the terms
  of the LGPL License and not to allow others to use your version of this file
  under the MPL, indicate your decision by deleting  the provisions above and
  replace  them with the notice and other provisions required by the LGPL
  License.  If you do not delete the provisions above, a recipient may use
  your version of this file under either the MPL or the LGPL License.

  For more information about the LGPL: http://www.gnu.org/copyleft/lesser.html

  rev. 1
  date 2014/11/06
  auth ICHIMARU takeshi ( http://homepage1.nifty.com/Ayakawa/ )
  memo original => Jpeg2000Bitmap.pas( PasOpenJpeg ver 1.03 & 1.20 )
       convert from opj_compress.c/opj_decompress.c

  rev. 2
  date 2014/11/07
  auth ICHIMARU takeshi ( http://homepage1.nifty.com/Ayakawa/ )
  memo some changes for openjpeg21.pas rev.2

  rev. 3
  date 2014/11/26
  auth ICHIMARU takeshi ( http://homepage1.nifty.com/Ayakawa/ )
  memo color profile
       bug fix : CMYKtoDib
                 experience : define J2K_COLOR_PROFILE when you want to use color profile
  rev. 4
  date 2014/11/28
  auth ICHIMARU takeshi ( http://homepage1.nifty.com/Ayakawa/ )
  memo add .jpf extention

  rev. 5
  date 2016/10/27
  auth ICHIMARU takeshi ( http://ayakawa.o.oo7.jp )
  memo support EYCC
       enable color profile

  rev. 6
  date 2017/08/11
  auth ICHIMARU takeshi ( http://ayakawa.o.oo7.jp )
  memo support multithread for v2.2.0

  rev. 7
  date 2017/08/13
  auth ICHIMARU takeshi ( http://ayakawa.o.oo7.jp )
  memo rewrite ICC Profile support
       fix alpha bug of 32bit color format

  rev. 8
  date 2017/08/14
  auth ICHIMARU takeshi ( http://ayakawa.o.oo7.jp )
  memo add .j2c extension

  rev. 9
  date 2023/07/02
  auth ICHIMARU takeshi ( http://ayakawa.o.oo7.jp )
  memo mask invalidop & overflow exception for WIN64
}

unit Jpeg2kBitmap;

interface

uses
  Windows, SysUtils, Classes, Graphics, Math;

type
  TJpeg2000QualityRange = 1..100;

  { TBitmap descendant which loads and saves raster data as JPEG 2000 images.
    It can load both files with JP2 header and raw code streams. Only files
    with JP2 header are saved. When saving bitmap as JPEG 2000 image
    user can choose whether to use lossy or lossless compression and
    quality of lossy compression.

    Images with more/less than 8 bits per channel are resampled to 8 bit. Images
    using YCC color space are converted to RGB. Grayscale JPEG 20000 images
    are loaded as 8bit indexed bitmaps and default grayscale palette is used.
    Images with 2 or 4+ channels are treated like grayscale and only first
    channel is loaded. JPEG 2000 images that use ICC profile
    or palette are not supported - raw pixel data is loaded though.

    Opacity/alpha channel is supported - 32bit bitmap is created if
    opacity component is present in file. When using Delphi 2009+
    TBitmap's alpha format property is properly set.}

////////////////////////////////////////////////////////////////////////////
//
// define J2K_USE_COLOR_PROFILE when you want to use ICC Profile
{$DEFINE J2K_USE_COLOR_PROFILE}
//
////////////////////////////////////////////////////////////////////////////

  TJpeg2000Bitmap = class(TBitmap)
  private
{$IFDEF J2K_USE_COLOR_PROFILE}
    FUseICC : boolean;
{$ENDIF}
    FThreadNum : nativeint;
    FLossless: Boolean;
    FQuality: TJpeg2000QualityRange;
  protected
    procedure SetThreadNum(ANum:nativeint);
  public
    constructor Create; override;

    procedure LoadFromStream(Stream: TStream); override;
    procedure SaveToStream(Stream: TStream); override;

    { If set to True bitmap is saved as lossless JPEG 2000, otherwise it
      uses lossy compression.}
    property LosslessCompression: Boolean read FLossless write FLossless;
    { Quality of lossy compression where 1 is ugly image/small file and
      100 is max quality/large file.}
    property CompressionQuality: TJpeg2000QualityRange read FQuality write FQuality;
    property ThreadNum: Nativeint read FThreadNum write SetThreadNum;
{$IFDEF J2K_USE_COLOR_PROFILE}
    property UseColorProfile: boolean read FUseICC write FUseICC;
{$ENDIF}
  end;

  EJpeg2000Error = class(EInvalidGraphic);

const
  SJpeg2000ImageFile = 'JPEG 2000 Image';
  Jpeg2000DefaultLosslessCompression = False;
  Jpeg2000DefaultLossyQuality = 80;

implementation

uses
  OpenJpeg21,icm20apis;

type
  { Type Jpeg 2000 file (needed for OpenJPEG codec settings).}
  TJpeg2000FileType = (jtInvalid, jtJP2, jtJ2K, jtJPT);

  TChar8 = array[0..7] of AnsiChar;
  TChar4 = array[0..3] of AnsiChar;

const
  JP2Signature: TChar8 = #0#0#0#$0C#$6A#$50#$20#$20;
  J2KSignature: TChar4 = #$FF#$4F#$FF#$51;

function _opjReadStream(p_buffer:pointer; p_nb_bytes:OPJ_SIZE_T; p_user_data:pointer):OPJ_SIZE_T; cdecl;
var
  pStream : TStream;
  pBuf : pByte;
begin
  pStream:=TStream(p_user_data);
  pBuf:=PByte(p_buffer);
  result:=pStream.Read(pBuf^,p_nb_bytes);
  if result=0 then result:=-1;
{$IFDEF DEBUG}
  OutputDebugString(PWideChar('read: '+IntToStr(p_nb_bytes)+' => '+IntToStr(result)));
{$ENDIF}
end;

function _opjWriteStream(p_buffer:pointer; p_nb_bytes:OPJ_SIZE_T; p_user_data:pointer):OPJ_SIZE_T; cdecl;
var
  pStream : TStream;
  pBuf : pByte;
begin
  pStream:=TStream(p_user_data);
  pBuf:=PByte(p_buffer);
  result:=pStream.Write(pBuf^,p_nb_bytes);
{$IFDEF DEBUG}
  OutputDebugString(PWideChar('write: '+IntToStr(p_nb_bytes)+' => '+IntToStr(result)));
{$ENDIF}
end;

function _opjSkipStream(p_nb_bytes:OPJ_OFF_T; p_user_data:pointer):OPJ_OFF_T; cdecl;
var
  pStream : TStream;
  orgpos,newpos : INT64;
begin
  pStream:=TStream(p_user_data);
  orgpos:=pStream.Position;
  newpos:=pStream.Seek(p_nb_bytes, soCurrent);
  if (orgpos+p_nb_bytes)=newpos then
    result:=p_nb_bytes
  else
    result:=-1;
{$IFDEF DEBUG}
  OutputDebugString(PWideChar('skip: '+IntToStr(p_nb_bytes)+ '=> '+IntToStr(result)));
{$ENDIF}
end;

function _opjSeekStream(p_nb_bytes:OPJ_OFF_T; p_user_data:pointer):OPJ_BOOL; cdecl;
var
  pStream : TStream;
  newpos : INT64;
begin
  pStream:=TStream(p_user_data);
  newpos:=pStream.Seek(p_nb_bytes, soBeginning);
  if newpos=p_nb_bytes then
    result:=OPJ_TRUE
  else
    result:=OPJ_FALSE;
{$IFDEF DEBUG}
  OutputDebugString(PWideChar('seek: '+IntToStr(p_nb_bytes)+' => '+IntToStr(newpos)));
{$ENDIF}
end;

{$IFDEF DEBUG}
procedure _opjInfo(msg:PAnsiChar; client_data:pointer); cdecl;
var
  m : string;
begin
  m:='INFO : '+msg;
  OutputDebugString(@m[1]);
end;

procedure _opjWarning(msg:PAnsiChar; client_data:pointer); cdecl;
var
  m : string;
begin
  m:='WARN : '+msg;
  OutputDebugString(@m[1]);
end;

procedure _opjError(msg:PAnsiChar; client_data:pointer); cdecl;
var
  m : string;
begin
  m:='ERR  : '+msg;
  OutputDebugString(@m[1]);
end;
{$ENDIF}

constructor TJpeg2000Bitmap.Create;
begin
  inherited Create;
  FLossless := Jpeg2000DefaultLosslessCompression;
  FQuality  := Jpeg2000DefaultLossyQuality;
  FThreadNum:= 0; // default: no multi thread
{$IFDEF J2K_USE_COLOR_PROFILE}
  FUseICC   := false;
{$ENDIF}
end;

procedure TJpeg2000Bitmap.SetThreadNum(ANum:nativeint);
begin
  if ANum<0 then
    FThreadNum:=0
  else
    FThreadNum:=ANum;
end;


procedure TJpeg2000Bitmap.LoadFromStream(Stream: TStream);
var
  parameters: opj_dparameters_t;
  p_codec: popj_codec_t;
  l_stream: popj_stream_t;
  pimage: popj_image_t;
  threadsup_ok : boolean;
  errmsg : string;
{$IFDEF WIN64}
  svmask : TSSEExceptionMask;
{$ENDIF}

  function GetFileType: TJpeg2000FileType;
  var
    Count: Integer;
    Id: TChar8;
  begin
    Result := jtInvalid;
    Count := Stream.Read(Id, SizeOf(Id));
    if Count = SizeOf(Id) then
    begin
      // Check if we have full JP2 file format or just J2K code stream
      if CompareMem(@Id, @JP2Signature, SizeOf(JP2Signature)) then
        Result := jtJP2
      else if CompareMem(@Id, @J2KSignature, SizeOf(J2KSignature)) then
        Result := jtJ2K;
    end;
    Stream.Seek(-Count, soFromCurrent);
  end;

  procedure CreateGrayPalette;
  var
    I: Integer;
    LogPalette: TMaxLogPalette;
  begin
    FillChar(LogPalette, SizeOf(LogPalette), 0);
    LogPalette.palVersion := $300;
    LogPalette.palNumEntries := 256;
    for I := 0 to 255 do
    with LogPalette do
    begin
      palPalEntry[I].peRed :=   I;
      palPalEntry[I].peGreen := I;
      palPalEntry[I].peBlue :=  I;
    end;
    Self.Palette := CreatePalette(PLogPalette(@LogPalette)^);
  end;

  procedure ImgToDib;
    var
      maxW,maxH,
      adjustR,adjustG,adjustB,adjustA : integer;
      idx,i,j : integer;
      flg : boolean;
    function ToByte(v:integer):byte; inline;
    begin
      if v>255 then
        result:=255
      else if v<0 then
        result:=0
      else
        result:=Byte(v);
    end;
    function ConvertData(data, adjust :integer; sgnd, prec:OPJ_UINT32):Byte;
    begin
      if sgnd<>0 then inc(data, 1 SHL (prec-1));
      result := ToByte((data SHR adjust) + ((data SHR (adjust-1)) mod 2));
    end;
    function CalcAdjust(prec:integer):integer; inline;
    begin
      if prec>8 then result:=prec-8 else result:=0;
    end;
    procedure GRAYtoDib;
    var
      h,w : integer;
      p : PByte;
    begin
      adjustR:=CalcAdjust(pimage.comps[0].prec);
      idx:=0;
      for h:=0 to (maxH-1) do begin
        p:=ScanLine[h];
        for w:=0 to (maxW-1) do begin
           p^:=ConvertData(pimage.comps[0].data[idx],adjustR,pimage.comps[0].sgnd,pimage.comps[0].prec);
           inc(idx);
           inc(p);
        end;
      end;
    end;
    procedure SRGBtoDib;
    var
      h,w : integer;
      p : PByte;
    begin
      adjustR:=CalcAdjust(pimage.comps[0].prec);
      adjustG:=CalcAdjust(pimage.comps[1].prec);
      adjustB:=CalcAdjust(pimage.comps[2].prec);
      if PixelFormat=pf32bit then adjustA:=CalcAdjust(pimage.comps[3].prec);
      idx:=0;
      for h:=0 to maxH-1 do begin
        p:=ScanLine[h];
        for w:=0 to maxW-1 do begin
           // blue
           p^:=ConvertData(pimage.comps[2].data[idx],adjustB,pimage.comps[2].sgnd,pimage.comps[2].prec);
           inc(p);
           // green
           p^:=ConvertData(pimage.comps[1].data[idx],adjustG,pimage.comps[1].sgnd,pimage.comps[1].prec);
           inc(p);
           // red
           p^:=ConvertData(pimage.comps[0].data[idx],adjustB,pimage.comps[0].sgnd,pimage.comps[0].prec);
           inc(p);
           // alpha if needed
           if PixelFormat=pf32bit then begin
             p^:=ConvertData(pimage.comps[3].data[idx],adjustA,pimage.comps[3].sgnd,pimage.comps[3].prec);
             inc(p);
           end;
           inc(idx);
        end;
      end;
    end;
    procedure SYCCtoDib;
      procedure YCCtoRGB(y,cb,cr,offset:integer; var r,g,b:integer);
      begin
        Dec(cb,offset); Dec(cr,offset);
        r := Round(y                + 1.40200 * cr);
        g := Round(y - 0.34414 * cb - 0.71414 * cr);
        b := Round(y + 1.77200 * cb);
      end;
    var
      h,w,r,g,b,yidx,cbidx,cridx,ofs : integer;
      p : PByte;
    begin
      adjustR:=CalcAdjust(pimage.comps[0].prec);
      adjustG:=adjustR; adjustB:=adjustR;
      ofs:=1 SHL (pimage.comps[0].prec-1);
      for h:=0 to maxH-1 do begin
        p:=ScanLine[h];
        for w:=0 to maxW-1 do begin
          yidx:=w+h*maxW;
          cbidx:=(w div pimage.comps[1].dx) + (h div pimage.comps[1].dy) * (maxW div pimage.comps[1].dx);
          cridx:=(w div pimage.comps[2].dx) + (h div pimage.comps[2].dy) * (maxW div pimage.comps[2].dx);
          YCCtoRGB(pimage.comps[0].data[yidx],
                   pimage.comps[1].data[cbidx],
                   pimage.comps[2].data[cridx],
                   ofs,r,g,b);
          p^:=ConvertData(b,adjustB,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
          p^:=ConvertData(g,adjustG,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
          p^:=ConvertData(r,adjustR,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
        end;
      end;
    end;
    procedure EYCCtoDib;
    var
      y,cb,cr,r,g,b,flip_value,max_value,adjust,adjustA : integer;
      w,h : uint;
      sign1,sign2 : boolean;
      p : pByte;
      function lim(org,max:integer):integer; inline;
      begin
        if org>max then
          result:=max
        else if org<0 then
          result:=0
        else
          result:=org;
      end;
    begin
      // precheck
      if (pimage^.numcomps<3) OR
         (pimage^.comps[0].dx<>pimage^.comps[1].dx) OR
         (pimage^.comps[0].dx<>pimage^.comps[2].dx) OR
         (pimage^.comps[0].dy<>pimage^.comps[1].dy) OR
         (pimage^.comps[0].dy<>pimage^.comps[2].dy) then
        raise EJpeg2000Error.Create('this type EYCC is not supported');
      //
      flip_value:=1 SHL (pimage^.comps[0].prec-1);
      max_value:=(1 SHL pimage^.comps[0].prec)-1;
      sign1:=pimage^.comps[1].sgnd=0;
      sign2:=pimage^.comps[2].sgnd=0;
      adjust:=CalcAdjust(pimage.comps[0].prec);
      if PixelFormat=pf32bit then adjustA:=CalcAdjust(pimage.comps[3].prec);
      i:=0;
      for h:=0 to pimage^.comps[0].h-1 do begin
        p:=ScanLine[h];
        for w:=0 to pimage^.comps[0].w-1 do begin
          y:=pimage^.comps[0].data[i];
          cb:=pimage^.comps[1].data[i];
          cr:=pimage^.comps[2].data[i];
          if sign1 then dec(cb,flip_value);
          if sign2 then dec(cr,flip_value);
          r:=lim(trunc(y-0.0000368*cb+1.40199*cr+0.5),max_value);
          g:=lim(trunc(1.0003*y+0.344125*cb-0.7141128*cr+0.5),max_value);
          b:=lim(trunc(0.999823*y+1.77204*cb-0.000008*cr+0.5),max_value);
          p^:=ConvertData(b,adjust,0,0); inc(p);
          p^:=ConvertData(g,adjust,0,0); inc(p);
          p^:=ConvertData(r,adjust,0,0); inc(p);
          // alpha if needed
          if PixelFormat=pf32bit then begin
            p^:=ConvertData(pimage.comps[3].data[idx],adjustA,pimage.comps[3].sgnd,pimage.comps[3].prec);
            inc(p);
          end;
          // next
          inc(i);
        end;
      end;
    end;
    procedure CMYKtoDib;
      procedure CMYKtoRGB(c,m,y,k,upr:integer; var r,g,b:integer);
      begin
        r := upr - ( c - ( c * k div upr ) + k );
        g := upr - ( m - ( m * k div upr ) + k );
        b := upr - ( y - ( y * k div upr ) + k );
      end;
    var
      h,w,r,g,b,upr : integer;
      p : PByte;
    begin
      adjustR:=CalcAdjust(pimage.comps[0].prec);
      adjustG:=adjustR; adjustB:=adjustR;
      upr := (1 SHL pimage.comps[0].prec)-1;
      idx:=0;
      for h:=0 to maxH-1 do begin
        p:=scanline[h];
        for w:=0 to maxW-1 do begin
          CMYKtoRGB(pimage.comps[0].data[idx],pimage.comps[1].data[idx],
                    pimage.comps[2].data[idx],pimage.comps[3].data[idx],
                    upr,r,g,b);
          p^:=ConvertData(b,adjustB,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
          p^:=ConvertData(g,adjustG,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
          p^:=ConvertData(r,adjustR,pimage.comps[0].sgnd,pimage.comps[0].prec); inc(p);
          Inc(idx);
        end;
      end;
    end;
  begin
    //
    if pimage.comps[0].prec < 8 then
      raise EJpeg2000Error.Create('Unsupported number of components : '+IntToStr(pimage.comps[0].prec));
    //
    maxW := pimage.comps[0].w; maxH := pimage.comps[0].h;
    if PixelFormat = pf8bit then
      GRAYtoDib
    else
      case pimage.color_space of
        OPJ_CLRSPC_UNKNOWN,
        OPJ_CLRSPC_UNSPECIFIED:
          if (pimage.numcomps<3) or (pimage.numcomps>4) then
            GRAYtoDib
          else begin
            flg:=false;
            for i:=0 to pimage.numcomps-2 do begin
              for j:=i+1 to pimage.numcomps-1 do begin
                if (pimage.comps[i].dx<>pimage.comps[j].dx) or
                   (pimage.comps[i].dy<>pimage.comps[j].dy) then begin
                  flg:=true; break; // YCC ?
                end;
              end;
            end;
            if flg then
              SYCCtoDib
            else
              SRGBtoDib;
          end;
        OPJ_CLRSPC_SRGB:
          SRGBtoDib;
        OPJ_CLRSPC_GRAY:
          GRAYtoDib;
        OPJ_CLRSPC_SYCC:
          SYCCtoDib;
        OPJ_CLRSPC_EYCC:
          EYCCtoDib;
        OPJ_CLRSPC_CMYK:
          CMYKtoDib;
        else
          raise EJpeg2000Error.Create('unsupported color space : '+IntToStr(ord(pimage.color_space)));
      end;
  end;

  // use icm2.0 functions
  procedure ApplyICCProfile3;
  var
    hPrfs : Array[0..1] of HPROFILE;
    hTrns : HTRANSFORM;
    // HPROFILEの正当性チェック
    function IsValidProf(hp:HPROFILE):boolean;
    var
      bValid : BOOL;
    begin
      bValid:=false;
      IsColorProfileValid(hp,bValid);
      result:=bValid;
    end;
    // ソースのHPROFILE作成
    function MakeSrcHP:boolean;
    var
      pd : TProfile;
    begin
      result:=false;
      with pd do begin
        dwType:=PROFILE_MEMBUFFER;
        pProfileData:=pimage.icc_profile_buf;
        cbDataSize:=pimage.icc_profile_len;
      end;
      hPrfs[0]:=OpenColorProfile(@pd,PROFILE_READ,FILE_SHARE_READ,OPEN_EXISTING);
      if hPrfs[0]=0 then begin
{$IFDEF DEBUG}
        OutputDebugString(PWideChar('Err:OpenColorProfile > '+IntToStr(GetLastError)));
{$ENDIF}
        exit;
      end;
      if IsValidProf(hPrfs[0]) then
        result:=true
      else begin
{$IFDEF DEBUG}
        OutputDebugString(PWideChar('invalid src color profile'));
{$ENDIF}
        CloseColorProfile(hPrfs[0]);
      end;
    end;
    // 変換先のHPROFILE作成
    function MakeDstHP:boolean;
    var
      _hdc : HDC;
      lpcbName : DWORD;
      lpszFilename : PWideChar;
      pd : TProfile;
    begin
      result:=false;
      _hdc:=GetDC(0);
      if _hdc=0 then exit;
      try
        lpcbName:=0;
        GetICMProfile(_hdc,@lpcbName,lpszFilename);
        if GetLastError=ERROR_INSUFFICIENT_BUFFER then begin
          try
            lpszFilename:=AllocMem(lpcbName*sizeof(WideChar));
            try
              if GetICMProfile(_hdc,@lpcbName,lpszFilename) then begin
                FillChar(pd,sizeof(pd),0);
                with pd do begin
                  dwType:=PROFILE_FILENAME;
                  pProfileData:=lpszFilename;
                  cbDataSize:=lpcbName*SizeOf(WideChar);
                end;
                hPrfs[1]:=OpenColorProfile(@pd,PROFILE_READ,FILE_SHARE_READ,OPEN_EXISTING);
                if hPrfs[1]=0 then begin
{$IFDEF DEBUG}
                  OutputDebugString(PWideChar('Err:OpenColorProfile > '+IntToStr(GetLastError)));
{$ENDIF}
                end else begin
                  if IsValidProf(hPrfs[1]) then
                    result:=true
                  else begin
{$IFDEF DEBUG}
                    OutputDebugString(PWideChar('invalid dst color profile'));
{$ENDIF}
                    CloseColorProfile(hPrfs[1]);
                  end;
                end;
              end else begin
{$IFDEF DEBUG}
                OutputDebugString(PWideChar('Err:GetICMProfile > '+IntToStr(GetLastError)));
{$ENDIF}
              end;
            finally
              FreeMem(lpszFilename);
            end;
          except
            on EOutOfMemory do
{$IFDEF DEBUG}
              OutputDebugString(PWideChar('Err:GetICMProfile > Out of memory'))
{$ENDIF}
              ;
          end;
        end else begin
{$IFDEF DEBUG}
          OutputDebugString(PWideChar('Err:GetICMProfile > '+IntToStr(GetLastError)));
{$ENDIF}
        end;
      finally
        ReleaseDC(0,_hdc);
      end;
    end;
    //
    function MakeHTrns:boolean;
    var
      dwIntent : DWORD;
    begin
      result:=false;
      dwIntent:=$FFFFFFFF;
      hTrns:=CreateMultiProfileTransform(@hPrfs[0],2,@dwIntent,1,NORMAL_MODE,INDEX_DONT_CARE);
      if hTrns=0 then begin
{$IFDEF DEBUG}
        OutputDebugString(PWideChar('Err:CreateMultiProfileTransform > '+IntToStr(GetLastError)));
{$ENDIF}
      end else
        result:=true;
    end;
    //
    procedure ConvertData;
    var
      work : TBitmap;
      pSrcBits,pDstBits : pointer;
      pSrc,pDst : PByte;
      a,b : nativeuint;
      bmInput,bmOutput :BMFORMAT;
      w,h : integer;
    begin
      // 変換先の作業用bitmap作成
      work:=TBitmap.Create;
      try
        // 変換元のサイズとPixelFormatを合わせる
        work.SetSize(Width,height);
        work.PixelFormat:=PixelFormat;
        // 変換元バッファの先頭アドレスを取得
        a:=NativeUint(ScanLine[0]); b:=NativeUInt(ScanLine[Height-1]);
        if a<b then pSrcBits:=ScanLine[0] else pSrcBits:=ScanLine[Height-1];
        // 変換先バッファの先頭アドレスを取得
        a:=nativeUint(work.ScanLine[0]); b:=nativeUInt(work.ScanLine[Height-1]);
        if a<b then pDstBits:=work.ScanLine[0] else pDstBits:=work.ScanLine[Height-1];
        // Bitmap形式の設定
        case PixelFormat of
          pf32bit:
            begin
              bmInput:=BM_xBGRQUADS; bmOutput:=bmInput;
            end;
          pf24bit:
            begin
              bmInput:=BM_BGRTRIPLETS; bmOutput:=bmInput;
            end;
          else begin
{$IFDEF DEBUG}
            OutputDebugString(PWideChar('Err:unexpected pixel format'));
{$ENDIF}
            bmInput:=BM_GRAY; // dummy param
          end;
        end;
        if bmInput<>BM_GRAY then begin
          // 変換
          if not TranslateBitmapBits(hTrns,
                                     pSrcBits,bmInput,width,height,0,
                                     pDstBits,bmOutput,0,
                                     0,0
                                    ) then begin
            // エラー発生
{$IFDEF DEBUG}
            OutputDebugString(PWideChar('Err:TranslateBitmapBits > '+IntToStr(GetLastError)));
{$ENDIF}
          end else begin
            // Alphaがあるならそれも設定
            if PixelFormat=pf32bit then begin
              work.AlphaFormat:=AlphaFormat;
              for h:=0 to Height-1 do begin
                pSrc:=Scanline[h]; pDst:=work.Scanline[h];
                for w:=0 to Width-1 do begin
                  inc(pSrc); inc(pSrc); inc(pSrc);
                  inc(pDst); inc(pDst); inc(pDst);
                  pDst^:=pSrc^;
                  inc(pSrc); inc(pDst);
                end;
              end;
            end;
            // 元データにcopy
            canvas.draw(0,0,work);
          end;
        end;
      finally
        // 作業用bitmap削除
        work.Free;
      end;
    end;
    //
  begin
    // 変換元のHPROFILE作成
    if not MakeSrcHP then exit;
    try
      // 変換先のHPROFILE作成
      if MakeDstHP then
        try
          // HTRANSFORM作成
          if MakeHTrns then
            try
              // データコンバート
              ConvertData;
            finally
              DeleteColorTransform(hTrns);
            end;
        finally
          CloseColorProfile(hPrfs[1]);
        end;
    finally
      CloseColorProfile(hPrfs[0]);
    end;
  end;
  //
begin
  // init
  pimage:=nil;
{$IFDEF WIN64}
  svmask:=SetSSEExceptionMask(GetSSEExceptionMask + [exInvalidOp,exOverflow]);

  try
{$ENDIF}
    try
      // Determine which codec to use
      case GetFileType of
        jtJP2: p_codec := opj_create_decompress(OPJ_CODEC_JP2);
        jtJ2K: p_codec := opj_create_decompress(OPJ_CODEC_J2K);
      else
        raise EJpeg2000Error.Create('Unknown JPEG 2000 file type');
      end;
      if p_codec=nil then
        raise EJpeg2000Error.Create('failed : opj_create_decompress');

      try
        //* set decoding parameters to default values */
        opj_set_default_decoder_parameters(@parameters);
        // make read stream
        l_stream:=opj_stream_default_create(OPJ_TRUE {* mean 'READ' *});
        if l_stream=nil then
          raise EJpeg2000Error.Create('failed : opj_stream_default_create');

        try
          // setup callback funcs
  {$IFDEF DEBUG}
          opj_set_info_handler(p_codec,_opjInfo,nil);
          opj_set_warning_handler(p_codec,_opjWarning,nil);
          opj_set_error_handler(p_codec,_opjError,nil);
  {$ELSE}
          opj_set_info_handler(p_codec,nil,nil);
          opj_set_warning_handler(p_codec,nil,nil);
          opj_set_error_handler(p_codec,nil,nil);
  {$ENDIF}
          opj_stream_set_user_data(l_stream,Pointer(stream),nil);
          opj_stream_set_user_data_length(l_stream, stream.Size);
          opj_stream_set_read_function(l_stream,_opjReadStream);
          opj_stream_set_skip_function(l_stream,_opjSkipStream);
          opj_stream_set_seek_function(l_stream,_opjSeekStream);

          //* Setup the decoder decoding parameters using user parameters */
          if (OPJ_FALSE = opj_setup_decoder(p_codec, @parameters)) then
            raise EJpeg2000Error.Create('failed : opj_setup_decoder');

          // setup multi threads param if need it
          threadsup_ok:=false;
          try
            threadsup_ok:=(OPJ_TRUE=opj_has_thread_support);
          except
            on EOpenJp2Error do ; // no 2.2.0 nor later...
          end;
          if threadsup_ok and (ThreadNum>=1) then begin
            if OPJ_TRUE<>opj_codec_set_threads(p_codec,ThreadNum) then begin
              errmsg:='failed : opj_codec_set_threads > threadnum='+IntToStr(ThreadNum);
  {$IFDEF DEBUG}
              OutputDebugString(PWideChar(errmsg));
  {$ENDIF}
              raise EJpeg2000Error.Create(errmsg);
            end;
          end;

          //* Read the main header of the codestream and if necessary the JP2 boxes*/
          if OPJ_FALSE = opj_read_header(l_stream, p_codec, @pimage) then begin
            opj_image_destroy(pimage);
            raise EJpeg2000Error.Create('failed : opj_read_header');
          end;

          try
            if OPJ_FALSE = parameters.nb_tile_to_decode then begin
              //* Optional if you want decode the entire image */
              if (OPJ_FALSE = opj_set_decode_area(p_codec, pimage,
                                                  OPJ_INT32(parameters.DA_x0),
                                                  OPJ_INT32(parameters.DA_y0),
                                                  OPJ_INT32(parameters.DA_x1),
                                                  OPJ_INT32(parameters.DA_y1))) then
                raise EJpeg2000Error.Create('failed : opj_set_decode_area');
              //* Get the decoded image */
              if (OPJ_FALSE=opj_decode(p_codec,l_stream,pimage)) OR
                 (OPJ_FALSE=opj_end_decompress(p_codec,l_stream)) then
                raise EJpeg2000Error.Create('failed : decode image');
            end else begin
              //* It is just here to illustrate how to use the resolution after set parameters */
              {* 省略!
                if OPJ_FALSE=opj_set_decoded_resolution_factor(l_codec, 5) then
                  raise EJpeg2000Error.Create('failed : set the resolution factor tile');
               *}
              if (OPJ_FALSE = opj_get_decoded_tile(p_codec, l_stream, pimage, parameters.tile_index)) then
                raise EJpeg2000Error.Create('failed : decode tile');
            end;

            //* Close the byte stream */
            opj_stream_destroy(l_stream);
            l_stream:=nil;

            // Determine pixel format according to components
            case pimage.numcomps of
              3: PixelFormat := pf24bit;
              4: if pimage.color_space = OPJ_CLRSPC_SRGB then
                   PixelFormat := pf32bit
                 else
                   PixelFormat := pf24bit; // Do YCC or CMYK has Alpha?
              else
                // 1, 2, or 5+ components: take just the first one and treat it like grayscale
                PixelFormat := pf8bit;
            end;

            // Create grayscale palette for 8bit images - palette must be ready before
            // storing pixel data to bitmap!
            if PixelFormat = pf8bit then
              CreateGrayPalette;

            // convert to DIB
            SetSize(pimage.x1 - pimage.x0, pimage.y1 - pimage.y0);

            ImgToDib;
  {$IFDEF J2K_USE_COLOR_PROFILE}
            if UseColorProfile and (pimage.icc_profile_buf<>nil) then begin
              // apply icc profile
              ApplyICCProfile3;
              // free icc_profile_buf : free(pimage.icc_profile_buf)
              // pimage.icc_profile_buf:=nil; pimage.icc_profile_len:=0;
            end;
  {$ENDIF}
            if PixelFormat = pf32bit then AlphaFormat := afDefined;

          finally
            // must be destoryed
            opj_image_destroy(pimage);
          end;
        finally
          // must be destroyed
          if l_stream<>nil then opj_stream_destroy(l_stream);
        end;
      finally
        // must be destroyed
        opj_destroy_codec(p_codec);
      end;

    except
      ON E:EExternalException DO
        raise EJpeg2000Error.Create('DLL or Function is not found');
    end;
{$IFDEF WIN64}
  finally
    SetSSEExceptionMask(svmask);
  end;
{$ENDIF}
end;

procedure TJpeg2000Bitmap.SaveToStream(Stream: TStream);
var
  work : TBitmap;
  numcmpts : OPJ_UINT32;
  cmptparms : array[0..3] of opj_image_cmptparm_t;
  i,h,w,idx : integer;
  pimage : popj_image_t;
  p : PByte;
  prms : opj_cparameters_t;
  pcodec : popj_codec_t;
  pstream : popj_stream_t;
  Rate, TargetSize: Single;
  NumDataItems: Integer;
begin
  try
    if PixelFormat in [pf24bit,pf32bit] then
      work:=self
    else begin
      work:=TBitmap.Create;
      work.assign(self);
      work.PixelFormat:=pf24bit;
    end;
    try
      if PixelFormat=pf32bit then
        numcmpts:=4
      else
        numcmpts:=3;
      for i:=0 to 3 do begin
        FillChar(cmptparms[i],sizeof(cmptparms[i]),0);
        with cmptparms[i] do begin
          dx:=1;
          dy:=1;
          w:=width;
          h:=height;
          prec:=8;
          bpp:=8;
          sgnd:=0;
        end;
      end;
      pimage:=opj_image_create(numcmpts,@cmptparms[0],OPJ_CLRSPC_SRGB);
      if pimage=nil then
        raise EJpeg2000Error.Create('fail : opj_image_create');
      try
        idx:=0;
        for h:=0 to Height-1 do begin
          p:=Scanline[h];
          for w:=0 to Width-1 do begin
            pimage^.comps[2].data[idx]:=p^; Inc(p);
            pimage^.comps[1].data[idx]:=p^; Inc(p);
            pimage^.comps[0].data[idx]:=p^; Inc(p);
            if numcmpts=4 then begin
               pimage^.comps[3].data[idx]:=p^; Inc(p);
            end;
            inc(idx);
          end;
        end;
        opj_set_default_encoder_parameters(@prms);
        with prms do begin
          cod_format := 1;
          numresolution := 6;
          tcp_numlayers := 1;
          cp_disto_alloc := 1;
          tcp_mct:=1;
          if FLossless then begin
            tcp_rates[0]:=0;
          end else begin
            // Quality -> Rate computation taken from ImageMagick
            Rate := 100.0 / Sqr(115 - FQuality);
            NumDataItems := Width * Height * numcmpts;
            TargetSize := (NumDataItems * Rate) + 550 + (numcmpts - 1) * 142;
            tcp_rates[0] := 1.0 / (TargetSize / NumDataItems);
          end;
        end;
        pimage^.x0:=prms.image_offset_x0;
        pimage^.y0:=prms.image_offset_y0;
        if pimage^.x0 = 0 then
          pimage^.x1:=width
        else
          pimage^.x1:=pimage^.x0+width;
        if pimage^.y0 = 0 then
          pimage^.y1:=height
        else
          pimage^.y1:=pimage^.y0+height;
        pcodec:=opj_create_compress(OPJ_CODEC_JP2);
        if pcodec=nil then
          raise EJpeg2000Error.Create('fail : opj_create_compress');
        try
          if OPJ_FALSE = opj_setup_encoder(pcodec,@prms,pimage) then
            raise EJpeg2000Error.Create('fail : opj_setup_encoder');
          pstream:=opj_stream_default_create(OPJ_FALSE { mean WRITE });
          if pstream=nil then
            raise EJpeg2000Error.Create('fail : opj_stream_default_create');
          try
{$IFDEF DEBUG}
            opj_set_info_handler(pcodec,_opjInfo,nil);
            opj_set_warning_handler(pcodec,_opjWarning,nil);
            opj_set_error_handler(pcodec,_opjError,nil);
{$ELSE}
            opj_set_info_handler(pcodec,nil,nil);
            opj_set_warning_handler(pcodec,nil,nil);
            opj_set_error_handler(pcodec,nil,nil);
{$ENDIF}
            opj_stream_set_user_data(pstream,Pointer(stream),nil);
            opj_stream_set_write_function(pstream,_opjWriteStream);
            opj_stream_set_skip_function(pstream,_opjSkipStream);
            opj_stream_set_seek_function(pstream,_opjSeekStream);
            if OPJ_TRUE <> opj_start_compress(pcodec,pimage,pstream) then
              raise EJpeg2000Error.Create('fail : opj_start_compress');
            if OPJ_TRUE <> opj_encode(pcodec,pstream) then
              raise EJpeg2000Error.Create('fail : opj_encode');
            if OPJ_TRUE <> opj_end_compress(pcodec,pstream) then
              raise EJpeg2000Error.Create('fail : opj_end_compress');
          finally
            opj_stream_destroy(pstream);
          end;
        finally
          opj_destroy_codec(pcodec);
        end;
      finally
        opj_image_destroy(pimage);
      end;
    finally
      if work<>self then work.Free;
    end;
  except
    ON E:EOpenJp2Error DO
      raise EJpeg2000Error.Create('DLL or Function is not found');
    ON E:EExternalException DO
      raise EJpeg2000Error.Create('DLL or Function is not found');
  end;
end;

initialization
  TPicture.RegisterFileFormat('jp2', SJpeg2000ImageFile, TJpeg2000Bitmap);
  TPicture.RegisterFileFormat('j2k', SJpeg2000ImageFile, TJpeg2000Bitmap);
  TPicture.RegisterFileFormat('jpc', SJpeg2000ImageFile, TJpeg2000Bitmap);
  TPicture.RegisterFileFormat('jpf', SJpeg2000ImageFile, TJpeg2000Bitmap);
  TPicture.RegisterFileFormat('j2c', SJpeg2000ImageFile, TJpeg2000Bitmap);

finalization
  TPicture.UnregisterGraphicClass(TJpeg2000Bitmap);

end.
